'use strict';

// You can use your favorite http client package to make REST calls, however, the node fetch API is pre-installed with the bots-node-sdk.
// Documentation can be found at https://www.npmjs.com/package/node-fetch
// Un-comment the next line if you want to make REST calls using node-fetch. 
// const fetch = require("node-fetch");

module.exports = {
  metadata: {
    name: 'oda.PizzaEEH',
    eventHandlerType: 'ResolveEntities'
  },
  handlers: {
    entity: {
      /**
      * Default message handler that includes acknowledgements when a bag item is updated
      * or a bag item value is provided while the user was prompted for another item
      */
      publishMessage: async (event, context) => {
        updatedItemsMessage(context);
        outOfOrderItemsMessage(context);
        context.addCandidateMessages();
      },
      /**
       * This handler is called when the composite bag entity is resolved
       */
      resolved: async (event, context) => { // eslint-disable-line no-unused-vars
        let pizzaDoughHasValue = context.getItemValue('PizzaDough') != null ? true : false;
        if (!pizzaDoughHasValue) {
          //if  no value is set for the pizza dough, set it to 'regular'
          context.setItemValue('PizzaDough', 'regular');
        }
      }
      // add more entity level event handlers here
    },
    items: {
      PizzaTopping: {
        publishDisambiguateMessage: async (event, context) => {
          context.getLogger().info("publishDisambiguateMessage");

          let messageModel = context.getMessageModel();

          //define a list of values to select from
          var postbacks = [];

          //iterate the values in the disambiguation list and create a list item for each
          for (let i = 0; i < event.disambiguationValues.length; i++) {

            //add a character like "a", "b", "c" and the label itself as a keyword
            let chr = String.fromCharCode(97 + i);
            let keywords = [event.disambiguationValues[i], chr];

            // set label as e.g. a. <label>, b. <label>
            let label = chr + ". " + event.disambiguationValues[i];

            //define a JSON message that will be sent to the bot when a user selects a list item.
            //The message invokes the setPizzaTopping custom event
            let postbackPayload = { "event": { "name": "setPizzaTopping", "properties": { "PizzaTopping": event.disambiguationValues[i] } } };

            //add postback action object to list of values
            postbacks.push(messageModel.postbackActionObject(label, null, postbackPayload, keywords));
          }
          //Define the disambiguation dialog prompt
          let prompt = "You selected " + event.disambiguationValues.length + " toppings. Unfortunately You can only have one. Please choose from the list.";

          let textMessage = messageModel.textConversationMessage(prompt, postbacks);
          context.addMessage(textMessage);
        }
      },
      PizzaSize: {
        publishPromptMessage: async (event, context) => {
          context.getLogger().info(context.getCurrentItem() + " : publishPromptMessage");
          let promptCount = event.promptCount;
          if (promptCount > 0) {
            //add error message
            let candidateMessages = context.getCandidateMessages();
            let errorMessage = "Your provided message '" + context.getUserInput() +
              "' is not a valid pizza size. "
            candidateMessages[0].text = errorMessage + candidateMessages[0].text;
            context.addCandidateMessages(candidateMessages);
          }
          else {
            context.addCandidateMessages(context.getCandidateMessages());
          }
        }
      },
      DeliveryTime: {
        validate: async (event, context) => {
          const monthNames = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];

          //create order date (will be created in UTC as milliseconds)
          let orderDate = new Date();
          let orderDateTimeInMs = orderDate.getTime();

          //access delivery hours, minutes and whether it is am or pm
          let hrs = event.newValue.hrs;
          let mins = event.newValue.mins;
          let ampm = event.newValue.hourFormat;

          //get timezoneOffset from user profile (channel)
          let timezoneOffset = context.getVariable('profile.timezoneOffset');

          //determine max delivery time in milliseconds
          let orderDay = orderDate.getDate();
          let orderMonth = monthNames[orderDate.getMonth()];
          let orderYear = orderDate.getFullYear();
          let maxDeliveryTimeInMilliseconds = (new Date(orderMonth + " " + orderYear + ", " + orderDay + " 8:30 pm")).getTime() + (timezoneOffset != null ? timezoneOffset : 0);

          //compose delivery date string from order date day, month, year and 
          //the hours and min information passed through the TIME entity. This
          //avoids people ordering pizza e.g. for tomorrow or yesterday
          let deliveryDateString = monthNames[orderDate.getMonth()] + " " +
            orderDate.getFullYear() + ", " +
            orderDate.getDate() + " " +
            hrs + ":" + mins + " " + ampm;

          //delivery date should consider timezoneOffset if available
          let deliveryDateInMilliseconds = (new Date(deliveryDateString)).getTime() + (timezoneOffset != null ? timezoneOffset : 0);

          //hours and minutes cannot both be 0
          if ((hrs == 0) && (mins == 0)) {
            let message = "Please enter a time. Pizzas are delivered the same day. Orders must be placed 60 minutes before delivery, but at the latest at 8:30 p.m.";
            context.addValidationError('DeliveryTime', message);
            return false;
          }

          //check if shop is closed for deliveries, which is for when order time plus pizza 
          //creation time is later than max delivery time 
          if ((orderDateTimeInMs + 3600000) > maxDeliveryTimeInMilliseconds) {
            let message = "Sorry, but I only accept orders until 8:30 p.m. All orders after 8.30 pm must be processed by an employee";
            context.addMessage(message, true)
            context.cancel();
            return false;
          }

          //delivery is for "pm" hours until 9:30
          if (ampm.toLowerCase() === 'am') {
            let message = "Sorry, we only deliver from 12 pm to 9.30 pm. Orders must be received 60 minutes before delivery, that is, at 8:30 pm at the latest. Please choose another time.";
            context.addValidationError('DeliveryTime', message);
            return false;
          }

          //is delivery time in future? If it is not, raise validation error.           
          if (deliveryDateInMilliseconds < orderDateTimeInMs) {
            let message = "The delivery time cannot be in the past. Please enter a time that is at least 60 minutes ahead of the current time.";
            context.addValidationError('DeliveryTime', message);
            return false;
          }

          //is requested delivery time outside delivery hours?
          if (deliveryDateInMilliseconds > maxDeliveryTimeInMilliseconds) {
            let message = "We deliver until 9:30 pm. Orders must be placed 60 minutes before delivery. Please try again.";
            context.addValidationError('DeliveryTime', message);
            return false;
          }

          //is delivery within 60 minutes? If so, ask user to change to later
          if (deliveryDateInMilliseconds < (orderDateTimeInMs + 3600000)) {
            let message = "Quality takes time. We need at least 60 minutes to create the pizza. Please provide a new delivery time";
            context.addValidationError('DeliveryTime', message);
            return false;
          }

          return true;
          return true;
        }
      }
    },
    custom: {
      setPizzaTopping: async (event, context) => {
        context.clearItemValue("PizzaTopping");
        context.setItemValue("PizzaTopping", event.PizzaTopping);
        //set a confirmation message. The boolean value true at its end sets the "keepTurn"
        //behavior. True means that the bot continues with the next bag item, not waiting 
        //for user input
        context.addMessage("Great. Now that we've got this sorted, let us continue.", true);
      }
    }
  }
};

/**
 * Helper function to show acknowledgement message when a bag item value is updated.
 */
function updatedItemsMessage(context) {
  if (context.getItemsUpdated().length > 0) {
    let message = "I have updated" + context.getItemsUpdated().map((item, i) => (i !== 0 ? " and the " : " the ") + item.toLowerCase() + " to " + context.getDisplayValue(item));
    context.addMessage(message);
  }
}

/**
 * Helper function to show acknowledgement message when a bag item value is provided when user was prompted for anther bag item.
 */
function outOfOrderItemsMessage(context) {
  if (context.getItemsMatchedOutOfOrder().length > 0) {
    let message = "I got" + context.getItemsMatchedOutOfOrder().map((item, i) => (i !== 0 ? " and the " : " the ") + item.toLowerCase() + " " + context.getDisplayValue(item));
    context.addMessage(message);
  }
}

